package cn.com.qmth.examcloud.core.oe.admin.api.controller.client;

import cn.com.qmth.examcloud.api.commons.enums.ExamType;
import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.FileUtil;
import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamRecordDataBeanConvert;
import cn.com.qmth.examcloud.core.oe.admin.api.bean.ExamRecordDataDomain;
import cn.com.qmth.examcloud.core.oe.admin.base.Constants;
import cn.com.qmth.examcloud.core.oe.admin.base.utils.Check;
import cn.com.qmth.examcloud.core.oe.admin.dao.ExamRecordDataRepo;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExamRecordDataEntity;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExamRecordPaperStructEntity;
import cn.com.qmth.examcloud.core.oe.admin.dao.entity.ExamRecordQuestionsEntity;
import cn.com.qmth.examcloud.core.oe.admin.service.*;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.*;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.examscore.ObjectiveScoreInfo;
import cn.com.qmth.examcloud.core.oe.admin.service.bean.uploadfile.FileInfo;
import cn.com.qmth.examcloud.reports.commons.bean.AdminOperateReport;
import cn.com.qmth.examcloud.reports.commons.util.ReportsUtil;
import cn.com.qmth.examcloud.support.cache.bean.ExamPropertyCacheBean;
import cn.com.qmth.examcloud.support.enums.ExamProperties;
import cn.com.qmth.examcloud.support.helper.ExamCacheTransferHelper;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import cn.com.qmth.examcloud.web.support.ControllerSupport;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

@Api(tags = "（客户端）考试过程相关接口")
@RestController
@RequestMapping("${$rmp.ctr.oe}/client/exam/process")
public class ExamProcessController extends ControllerSupport {

    private static final Logger log = LoggerFactory.getLogger(ExamProcessController.class);

    @Autowired
    private ExamStudentService examStudentService;

    @Autowired
    private PracticeService practiceService;

    @Autowired
    private ExamRecordDataSyncService examRecordDataSyncService;

    @Autowired
    private ExamRecordPaperStructService examRecordPaperStructService;

    @Autowired
    private ExamRecordQuestionsService examRecordQuestionsService;

    @Autowired
    private ExamRecordDataRepo examRecordDataRepo;

    @Autowired
    private ExamScoreService examScoreService;

    @Autowired
    private OfflineExamService offlineExamService;

    @ApiOperation(value = "获取在线考试待考列表")
    @PostMapping("/queryExamList")
    public List<OnHandExamInfo> queryExamList() {
        User user = getAccessUser();
        return examStudentService.queryOnlineExamList(user.getUserId(), ExamType.ONLINE);
    }

    @ApiOperation(value = "获取在线考试已结束列表")
    @PostMapping("/queryExamEndList")
    public List<OnHandExamInfo> queryExamEndList() {
        User user = getAccessUser();
        return examStudentService.queryOnlineExamEndList(user.getUserId(), ExamType.ONLINE);
    }

    @ApiOperation(value = "获取在线作业待考列表")
    @PostMapping("/queryHomeworkList")
    public List<OnHandExamInfo> queryHomeworkList() {
        User user = getAccessUser();
        return examStudentService.queryOnlineExamList(user.getUserId(), ExamType.ONLINE_HOMEWORK);
    }

    @ApiOperation(value = "练习课程列表")
    @PostMapping("/queryPracticeCourseList")
    public List<PracticeCourseInfo> queryPracticeCourseList(@RequestParam Long examId) {
        User user = getAccessUser();
        return practiceService.queryPracticeCourseList(examId, user.getUserId());
    }

    @ApiOperation(value = "课程练习记录详情")
    @PostMapping("/queryPracticeRecordList")
    public List<PracticeRecordInfo> queryPracticeRecordList(@RequestParam Long examStudentId) {
        return practiceService.queryPracticeRecordList(examStudentId);
    }

    @ApiOperation(value = "单次练习答题情况统计")
    @PostMapping("/getPracticeDetailInfo")
    public PracticeDetailInfo getPracticeDetailInfo(@RequestParam Long examRecordDataId, @RequestParam(required = false) String fromCache) {
        if (fromCache != null) {
            examRecordDataId = examRecordDataSyncService.getExamRecordDataIdByCacheId(examRecordDataId);
            if (examRecordDataId == null) {
                throw new StatusException("1001", "未找到数据");
            }
        }
        return practiceService.getPracticeDetailInfo(examRecordDataId);
    }

    @ApiOperation(value = "获取考试记录试卷结构")
    @PostMapping("/getExamRecordPaperStruct")
    public ExamRecordPaperStructEntity getExamRecordPaperStruct(@RequestParam Long examRecordDataId, @RequestParam(required = false) String fromCache) {
        if (fromCache != null) {
            examRecordDataId = examRecordDataSyncService.getExamRecordDataIdByCacheId(examRecordDataId);
            if (examRecordDataId == null) {
                throw new StatusException("1001", "未找到数据");
            }
        }
        return examRecordPaperStructService.getExamRecordPaperStruct(examRecordDataId);
    }

    @ApiOperation(value = "获取交卷之后的所有试题作答信息")
    @PostMapping("/getExamRecordQuestions")
    public ExamRecordQuestionsEntity getExamRecordQuestions(@RequestParam Long examRecordDataId, @RequestParam(required = false) String fromCache) {
        if (fromCache != null) {
            examRecordDataId = examRecordDataSyncService.getExamRecordDataIdByCacheId(examRecordDataId);
            if (examRecordDataId == null) {
                throw new StatusException("1001", "未找到数据");
            }
        }
        return examRecordQuestionsService.getExamRecordQuestionsAndFixExamRecordDataIfNecessary(examRecordDataId);
    }

    @ApiOperation(value = "获取当前考试记录信息")
    @PostMapping("/findExamRecordDataEntity")
    public ExamRecordDataDomain findExamRecordDataEntity(@RequestParam Long examRecordDataId, @RequestParam(required = false) String fromCache) {
        if (fromCache != null) {
            examRecordDataId = examRecordDataSyncService.getExamRecordDataIdByCacheId(examRecordDataId);
            if (examRecordDataId == null) {
                throw new StatusException("1001", "未找到数据");
            }
        }
        ExamRecordDataEntity examRecordDataEntity = GlobalHelper.getEntity(examRecordDataRepo, examRecordDataId, ExamRecordDataEntity.class);
        ExamRecordDataDomain domain = ExamRecordDataBeanConvert.copyFrom(examRecordDataEntity);
        domain.setExamRecord(examRecordDataEntity);
        return domain;
    }

    @ApiOperation(value = "根据examStudentId获取客观分信息")
    @PostMapping("/queryObjectiveScoreList")
    public List<ObjectiveScoreInfo> queryObjectiveScoreList(@RequestParam Long examStudentId) {
        return examScoreService.queryObjectiveScoreList(examStudentId);
    }

    @ApiOperation(value = "获取离线考试列表")
    @PostMapping("/getOfflineCourse")
    public List<OfflineExamCourseInfo> getOfflineCourse() {
        User user = getAccessUser();
        return offlineExamService.getOfflineCourse(user.getUserId());
    }

    @ApiOperation(value = "离线考试:开始考试")
    @PostMapping("/startOfflineExam")
    public void startOfflineExam(@RequestParam long examStudentId) {
        offlineExamService.startOfflineExam(examStudentId);
    }

    @ApiOperation(value = "离线考试:交卷")
    @PostMapping("/submitPaper")
    public void submitPaper(@RequestParam Long examRecordDataId,
                            @RequestPart(value = "file", required = false) MultipartFile file) throws Exception {
        Check.isNull(file, "上传文件不能为空");
        User user = getAccessUser();

        String fileSuffix = FileUtil.getFileSuffix(file.getOriginalFilename());

        String fileType;
        if (".zip".equals(fileSuffix)) {
            fileType = "ZIP";
        } else if (".pdf".equals(fileSuffix)) {
            fileType = "PDF";
        } else if (".jpg.jpeg.png".contains(fileSuffix)) {
            fileType = "IMAGE";
        } else {
            fileType = "UNKNOWN";
        }

        FileInfo fileInfo = new FileInfo();
        fileInfo.setOriginalFileName(file.getOriginalFilename());
        fileInfo.setFileBytes(file.getBytes());
        fileInfo.setFileSize(file.getSize());
        fileInfo.setFileSuffix(fileSuffix);
        fileInfo.setFileType(fileType);

        offlineExamService.submitPaper(examRecordDataId, fileInfo, user.getUserId());
    }

    @ApiOperation(value = "离线考试:多文件交卷")
    @PostMapping("/batchSubmitPaper")
    public void batchSubmitPaper(@ApiParam(value = "考试记录id") @RequestParam Long examRecordDataId,
                                 @ApiParam(value = "文件类型:ZIP、PDF、IMAGE") @RequestParam String fileType,
                                 @ApiParam(value = "文件MD5信息") @RequestParam String[] fileMd5Array,
                                 @ApiParam(value = "文件数组") @RequestPart MultipartFile[] fileArray) throws Exception {
        Check.isNull(fileArray, "上传文件不能为空");

        if (fileArray.length != fileMd5Array.length) {
            throw new StatusException("400401", "文件数量和文件MD5数量不匹配");
        }

        Optional<ExamRecordDataEntity> optional = examRecordDataRepo.findById(examRecordDataId);
        if (!optional.isPresent()) {
            throw new StatusException("400403", "考试记录不存在");
        }

        ExamRecordDataEntity examRecordData = optional.get();
        if (ExamType.OFFLINE != examRecordData.getExamType()) {
            throw new StatusException("400403", "非离线考试");
        }

        if (!examStudentService.isEnableExamStudent(examRecordData.getExamStudentId())) {
            throw new StatusException("400403", "当前考生已禁用");
        }

        // 当前考试允许的附件类型，示例：["ZIP","PDF","IMAGE"]
        ExamPropertyCacheBean examProperty = ExamCacheTransferHelper.getCachedExamProperty(examRecordData.getExamId(),
                examRecordData.getStudentId(), ExamProperties.OFFLINE_UPLOAD_FILE_TYPE.name());
        String[] settingFileTypes = new JsonMapper().toArray(examProperty.getValue(), String.class);
        if (ArrayUtils.isEmpty(settingFileTypes)) {
            throw new StatusException("400403", "当前考试设置不允许上传附件");
        }

        String uploadFileType = fileType.toUpperCase();
        if (!ArrayUtils.contains(settingFileTypes, uploadFileType)) {
            throw new StatusException("400403", "当前考试允许上传文件格式为:" + examProperty.getValue());
        }

        if (("ZIP".equals(uploadFileType) || "PDF".equals(uploadFileType)) && fileArray.length > 1) {
            throw new StatusException("400403", "上传" + uploadFileType + "类型文件时只能传一个文件");
        }

        List<byte[]> resultFiles = new ArrayList<>();
        for (int i = 0; i < fileArray.length; i++) {
            MultipartFile file = fileArray[i];
            log.info("OfflineExam file {} - {}", i + 1, file.getOriginalFilename());

            String fileSuffix = FileUtil.getFileSuffix(file.getOriginalFilename());
            if (!matchFileTypes(new String[]{uploadFileType}, fileSuffix)) {
                throw new StatusException("400401", "文件格式不正确");
            }

            if (file.getSize() > Constants.ANSWER_FILE_MAX_SIZE * Constants.M_SIZE) {
                throw new StatusException("400402", "文件大小不能超过" + Constants.ANSWER_FILE_MAX_SIZE + "M");
            }

            String realMD5 = DigestUtils.md5Hex(file.getInputStream());
            String tempMD5 = fileMd5Array[i];
            if (!realMD5.equals(tempMD5)) {
                throw new StatusException("400403", "文件数据不完整，请重新上传");
            }

            resultFiles.add(file.getBytes());
        }

        // todo
        // offlineExamService.batchSubmitPaper(examRecordDataId, resultFiles, fileType);

        User user = getAccessUser();
        ReportsUtil.report(new AdminOperateReport(user.getRootOrgId(), user.getUserId(), "考试进度详情-上传作答",
                "考试记录ID:" + examRecordDataId + " 考生ID:" + examRecordData.getExamStudentId()));
    }

    private boolean matchFileTypes(String[] settingFileTypes, String fileSuffix) {
        if (ArrayUtils.isEmpty(settingFileTypes) || StringUtils.isEmpty(fileSuffix)) {
            return false;
        }

        for (String fileType : settingFileTypes) {
            if ("IMAGE".equals(fileType)) {
                if (".jpg.jpeg.png".contains(fileSuffix)) {
                    return true;
                }
            } else {
                // ZIP、PDF
                if (("." + fileType).equalsIgnoreCase(fileSuffix)) {
                    return true;
                }
            }
        }
        return false;
    }

}
